/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive;

import org.xmloperator.reactive.exception.DoneMachineException;
import org.xmloperator.reactive.exception.PrivilegiedAccessException;
import org.xmloperator.reactive.exception.AgentAlreadyLaunchedException;

/**
 * The task of a ReactiveMachine consists in launching agents and accessing
 *   communities.
 * Methods of this Object must accept to be called from multiple Threads. 
 */
public interface ReactiveMachine extends Machine {

	/**
	 * Agent profiles.
	 * They can be combined by concatenation.
	 */
	public static final String PROFILE_SENDER_CERTIFIED_MESSAGE_GENERATION = "S";
	public static final String PROFILE_ACTION_PROPOSITION = "P";
	
	public static final String PROFILE_COMMUNITY_CONNECTION = "A";
	public static final String PROFILE_EVENT_GENERATION = "E";
	public static final String PROFILE_UNCERTIFIED_MESSAGE_GENERATION = "U";
	
	public static final String PROFILE_COMMUNITY_CREATION = "C";
	public static final String PROFILE_AGENT_LAUNCHING = "L";
	public static final String PROFILE_ACTION_EXECUTION = "X";
	public static final String PROFILE_AGENT_PREVENTING = "V";
	
	public static final String PROFILE_RESTRICTED = "SP";
	public static final String PROFILE_STANDARD = "SPAEU";
	public static final String PROFILE_FULL = "SPAEUCLXV";
	
	/**
	 * Launches an Agent.
	 * @param agentLaunchingKey a key that make the start() method
	 *   usage exclusive to the ReactiveMachine.
	 * @param agent : an Agent
	 * @param context a launching context that will be passed to the Agent.
	 *   It can be null.
	 * @param agentProfile Agent profile. It is obtained by concatenation
	 *   of the PRO_* constants.
	 * @return the entityId.
	 * @exception PrivilegiedAccessException if the agentLaunchingKey value
	 *   is not the good one.
	 * @exception AgentAlreadyLaunchedException if the given Agent is already
	 *   running.
	 * @exception DoneMachineException throwed if this ReactiveMachine is done.
	 */
	public String launchAgent(
			long agentLaunchingKey,
			Agent agent,
			LaunchingContext context,
			String agentProfile)
		throws
			PrivilegiedAccessException,
			AgentAlreadyLaunchedException,
			DoneMachineException;
	
	/**
	 * Generates and returns a new key.
	 * Each key is different of zero.
	 * @return a new generated key.
	 */
	public long newKey();

	/**
	 * Returns the profile of the Agent with the given entityId.
	 * It is null if the Agent is not launched or has terminated.
	 * The profile of an Agent is a constant. Prevention doesn't change
	 *   agent profile.
	 * @param entityId the entityId of an Agent.
	 * @return the profile of the Agent with the given entityId.
	 */
	public String getAgentProfile(String entityId);
	
	/**
	 * Prevents the Agent with the given entityId to operate on
	 *   this ReactiveMachine.
	 * @param agentPreventingKey a key that restricts the hability
	 *   for preventing an Agent to operate on the reactive machine.
	 * @param entityId the entityId of an Agent to be prevented to operate
	 *   on this ReactiveMachine.
	 * @param concernsAnyAccess true for preventing any access,
	 *   including suspend() and stop(), false for preventing only operations
	 *   such as generating events.
	 * @exception PrivilegiedAccessException if the agentPreventingKey value
	 *   is not the good one.
	 * @exception DoneMachineException throwed if this ReactiveMachine is done.
	 */
	public void preventAgent(
			long agentPreventingKey,
			String entityId,
			boolean concernsAnyAccess)
		throws PrivilegiedAccessException, DoneMachineException;
	
	/**
	 * Returns the ReactiveCommunity attached to a given thema.
	 * Returns null if no ReactiveCommunity is attached to the given thema.
	 * @param thema a thema.
	 * @return the ReactiveCommunity attached to a given thema.
	 */
	public ReactiveCommunity getCommunity(String thema);
	
	/**
	 * Waits for all agents terminate and all communities be closed,
	 *   then terminates this ReactiveMachine.
	 * @param machineDoneKey a key that restricts the usage of this method.
	 * @exception PrivilegiedAccessException if the machineDoneKey value
	 *   is not the good one.
	 */
	public void done(long machineDoneKey) throws PrivilegiedAccessException;
}
