/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.element;

/**
 * An element like in W3C-DOM.
 * See DocumentWriter for writing an XML document from a ParentElement.
 * An Element is either a DataElement, a ParentElement or an empty Element.
 * After reading as XML, an Element is either DataElement or ParentElement.
 *   An empty Element becomes a ParentElement with no child.
 */
public interface Element {
	
	public static final String REACTIVE_NAMESPACE =
		"http://www.xmloperator.net/reactive/namespace";
	
	/**
	 * Returns true if this Element is a DataElement, false elsewhere.
	 * @return true if this Element is a DataElement, false elsewhere.
	 */
	public boolean isDataElement();
	
	/**
	 * Returns true if this Element is a ParentElement, false elsewhere.
	 * @return true if this Element is a ParentElement, false elsewhere.
	 */
	public boolean isParentElement();
	
	/**
	 * Returns the unqualified local name of this Element;
	 * @return the unqualified local name of this Element;
	 */	
    public String getLocalName();

	/**
	 * Returns the namespace URI associated to this Element.
	 * It is internalized. There is no need to use the equals() method
	 *   for comparing it with an internalized value.
	 * @return the namespace URI associated to this Element.
	 */
    public String getNamespaceURI();

	/**
	 * Returns the parent of this Element if any, null elsewhere.
	 * A root Element has no parent.
	 * @return the parent of this Element if any, null elsewhere.
	 */
    public ParentElement getParent();
    
    /**
     * Returns the root Element of the hierarchy that contains this Element.
     * @return the root Element of the hierarchy that contains this Element.
     */
    public Element root();

	/**
	 * Returns the previous sibling of this Element if any, null elsewhere.
	 * @return the previous sibling of this Element if any, null elsewhere.
	 */
    public Element getPreviousSibling();

	/**
	 * Returns the next sibling of this Element if any, null elsewhere.
	 * @return the next sibling of this Element if any, null elsewhere.
	 */
    public Element getNextSibling();
    
    /**
     * Returns true if this Element is equivalent to the given one,
     *   false elsewhere.
     * Parents of the two elements are not taken into account.
     * An Element with children elements is equivalent to the same Element
     *   with equivalent children elements.
     * An Element with data is equivalent to the same Element with the same data.
     * @param element an Element.
     * @return true if this Element is equivalent to the given one,
     *   false elsewhere.
     */
    public boolean equals(Element element);
    
	/**
	 * Returns the count of previous sibling.
	 * @return the count of previous sibling.
	 */
	public int previousSiblingCount();
    
	/**
	 * Returns the hierarchical position of this Element.
	 * This is [] for the root.
	 * This is "[.1]" for the first child of the root.
	 * This is "[.1.2]" for the second child of the first child of the root.
	 * @return the hierarchical position of this Element.
	 */
	public String hierarchicalPosition();
}
