/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.ReactiveEntity;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.exception.ReactiveException;

/**
 * Event implementation.
 */
class EventImpl implements Event {

	private static int eventCount = 0;
	
	protected final String id = String.valueOf(++EventImpl.eventCount);
	protected boolean isValid = true;
	protected final ReactiveCommunity community;
	protected ReactiveEntity owner;
	/*
	 * In absence of a ReactiveEntity owner, the ReactiveCommunity is considered
	 *   as the owner.
	 */
	protected String name = null;
	protected long generationInstant = 0;

	EventImpl(final ReactiveCommunity community) {
		this(community, null);
	}
	
	EventImpl(final ReactiveCommunity community, final ReactiveEntity owner) {
		if (community == null)
			throw new ReactiveException(
				"The ReactiveCommunity of a valid Event cannot be null.");
		if (owner != null && owner.getCommunity() != community)
			throw new ReactiveException(
"The ReactiveCommunity of the owner of an Event must be the same than the one of the Event.");
		this.community = community;
		this.owner = owner;
	}

	protected final void verifyValidity() {
		if (!this.isValid)
			throw new ReactiveException("This Event is no more valid.");
	}
	
	final ReactiveEntity getOwner() {
		return this.owner;
	}
	
	final void setName(String name) {
		if (name == null)
			throw new IllegalArgumentException();
		this.verifyValidity();
		this.name = name.intern();
		this.owner = null;
	}

	void unvalidate() {
		this.isValid = false;
		this.name = null;
		this.owner = null;
	}

	/**
	 * Generates this Event at the given instant.
	 * @param instant : the last instant of occuring of this Event.  
	 * @return <tt>true</tt> if this is a new generation.
	 */
	final boolean generate(final long instant) {
		this.verifyValidity();
		if (instant < this.generationInstant)
			throw new ReactiveException("Going to the past is forbiden.");
		boolean isNew;
		synchronized (this) {
			isNew = this.generationInstant != instant;
			this.generationInstant = instant;
		}
		return isNew;
	}

	////////////////////////
	// Object implementation
	////////////////////////

	public String toString() {
		return "E" + this.id;
	}
	
	///////////////////////
	// Event implementation
	///////////////////////
	
	public final boolean isValid() {
		return this.isValid;
	}
	
	public final ReactiveCommunity getCommunity() {
		return this.community;
	}

	public String getName() {
		return this.name;
	}
	
	public boolean isBaseEvent() {
		return true;
	}
	
	public boolean isMessage() {
		return false;
	}
}
