/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import java.io.PrintStream;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.xmloperator.reactive.exception.BoardAlreadyAttachedException;
import org.xmloperator.reactive.exception.BoardAttachmentException;
import org.xmloperator.reactive.exception.ClosedCommunityException;
import org.xmloperator.reactive.exception.CommunityOwnerException;
import org.xmloperator.reactive.exception.NameAlreadyUsedException;
import org.xmloperator.reactive.exception.ReactiveException;
import org.xmloperator.reactive.exception.UndefinedEventOrMessageNameException;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.Board;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveEntity;

/**
 * ReactiveCommunity implementation.
 */
final class ReactiveCommunityImpl extends Thread implements ReactiveCommunity {
	
	//////////////
	// Static data
	//////////////
		
	private static int reactiveCommunityCount = 0;
	private static PrintStream trace = null;
	private static boolean isFullTrace = false;
	
	/**
	 * The automata status when no ReactiveEntity is connected.
	 */
	private static final int STATUS_IDLE = 0;

	/**
	 * The status of a transition before STATUS_ACTIVE.
	 * This transition may be used to connect a ReactiveEntity
	 * 	but not more than one.
	 */
	private static final int STATUS_BEFORE_ACTIVE = 1;
	
	/**
	 * Thread are active only during this period of time.
	 * It is closed when no more thread is active.
	 */
	private static final int STATUS_ACTIVE = 2;
	
	/**
	 * The status of an interruption of STATUS_ACTIVE.
	 */
	private static final int STATUS_SUSPENDED = 3;
	
	/**
	 * Instant is incremented during this transition.
	 * It is followed by BEFORE_ACTIVE or BEFORE_IDLE.
	 */
	private static final int STATUS_INSTANT_TRANSITION = 4;
	
	/////////////////
	// Static methods
	/////////////////

	static void setTrace(PrintStream trace, boolean isFullTrace) {
		ReactiveCommunityImpl.trace = trace;
		ReactiveCommunityImpl.isFullTrace = isFullTrace;
	}
		
	static String arrayToString(Object[] array) {
		StringBuffer sb = new StringBuffer();
		for (int k = 0; k < array.length; k++) {
			if (k != 0)
				sb.append(" ");
			sb.append(array[k]); 
		}
		return sb.toString();
	}

	/////////////////////////////
	// Shared final instance data
	/////////////////////////////
	
	final int id = ++ReactiveCommunityImpl.reactiveCommunityCount;
	final String communityId = "C" + String.valueOf(this.id);
	final ReactiveMachineImpl reactiveMachine;
	final ReactiveEntity creatorEntity;
	final MessageImpl actionMessage;
	final String theme;

	////////////////////////
	// Private instance data
	////////////////////////
	
	private boolean isClosed = false;
	private int status = STATUS_IDLE;
	private long instant = 0;
	private boolean instantHasBeenIdle = true;
	private int instantBroadcastCount = 0;
	private int instantBroadcastedEventCount = 0;
	private boolean isAutomataWaiting = false;
	private boolean isAutomataNotified = false;
	private final Object synchroAutomataActionRequired = new Object();
	private final Object synchroInactivity = new Object();
	private final Object synchroStatus = new Object();
	private final Object synchroNewInstantActive = new Object();
	private final Object synchroReleaseSuspend = new Object();
	private final Object synchroEventOrConnection = new Object();
	private final List entityToConnectList = new ArrayList();
	private final Set connectedEntitySet = new HashSet();
	private int activeEntityCount = 0;
	private int suspendedEntityCount = 0;
	private int waitingSuspendedEntityCount = 0;
	private final List broadcastInstantaneousEventList = new ArrayList();
	private final List broadcastDeferredEventList = new ArrayList();
	private boolean isActionMessageGenerated = false;
	private Event[] broadcastEvents = null;
	private Event[][] eventss = new Event[10][];
	private final List eventList = new ArrayList();
	private final Map fromNameToEvent = new HashMap();
	private ReactiveReporter reporter = null;
	private Board futureBoard = null;
	private Board board = null;
	private long boardWritingKey = 0;

	//////////////
	// Constructor
	//////////////
	
	ReactiveCommunityImpl(ReactiveMachineImpl reactiveMachine,
			ReactiveEntity creatorEntity, String theme) {
		super();
		if (reactiveMachine == null || creatorEntity == null)
			throw new IllegalArgumentException();
		this.reactiveMachine = reactiveMachine;
		this.creatorEntity = creatorEntity;
		this.theme = theme;
		this.actionMessage = new MessageImpl(this, null, true, true);
		for (int k = 0; k < this.eventss.length; k++)
			this.eventss[k] = new Event[k];
		this.eventList.add(this.actionMessage);
		this.setName(this.communityId);
		this.start();
		Thread.yield();
	}

	////////////////////
	// Statistic support
	////////////////////
	
	void setReporter(ReactiveReporter reporter) {
		this.reporter = reporter;
	}
	
	////////////////////////
	// Thread implementation
	////////////////////////
		
	public void run() {
		this.reactiveMachine.notifyCommunityStarting(this);
		while(true) {
			if (this.isAutomataWaiting || this.isAutomataNotified)
				throw new ReactiveException("Automata is not synchronized");
			try {
				synchronized (this.synchroAutomataActionRequired) {
					this.isAutomataWaiting = true;
					this.synchroAutomataActionRequired.wait();
				}
			}
			catch (InterruptedException exception) {}
			this.isAutomataWaiting = false;
			this.isAutomataNotified = false;
			if (this.status != STATUS_ACTIVE) {
				if (this.isClosed) {
					this.reactiveMachine.notifyCommunityTermination(this);
					return;
				}
				throw new ReactiveException(
						"Automata status should be ACTIVE.");
			}
			this.status = STATUS_SUSPENDED;
			if ((this.suspendedEntityCount != 0)
					&& !this.broadcastInstantaneousEventList.isEmpty()) {
				int broadcastEventCount = this.broadcastInstantaneousEventList.size(); 
				this.broadcastEvents =
					(broadcastEventCount < this.eventss.length)
					? this.eventss[broadcastEventCount]
					: new Event[broadcastEventCount];
				this.broadcastInstantaneousEventList.toArray(this.broadcastEvents);
				this.broadcastInstantaneousEventList.clear();
				this.instantBroadcastCount += 1;
				this.instantBroadcastedEventCount +=
					this.broadcastEvents.length;
				if (trace != null)
					trace.println(this + " broadcast "
						+ arrayToString(this.broadcastEvents));
				while (this.waitingSuspendedEntityCount
						!= this.suspendedEntityCount)
					Thread.yield();
				this.activeEntityCount = this.suspendedEntityCount;
				this.waitingSuspendedEntityCount = 0;
				this.suspendedEntityCount = 0;
				synchronized (this.synchroReleaseSuspend) {
					this.synchroReleaseSuspend.notifyAll();
					this.status = STATUS_ACTIVE;
				}
				continue;
			}
			this.broadcastEvents = null;
			this.broadcastInstantaneousEventList.clear();
			if (this.reporter != null)
				this.reporter.reportInstantEnd(
					this.instantHasBeenIdle,
					this.instantBroadcastCount,
					this.instantBroadcastedEventCount);
			this.status = STATUS_INSTANT_TRANSITION;
			this.instant += 1;
			if (trace != null)
				trace.println(this + " instant " + this.instant);
			this.status = STATUS_BEFORE_ACTIVE;
			if (this.reporter != null)
				this.reporter.reportInstantBegin(this.instant);
			this.instantBroadcastCount = 0;
			this.instantBroadcastedEventCount = 0;
			while (!this.entityToConnectList.isEmpty()) {
				ReactiveEntity reactiveEntity;
				synchronized (this.entityToConnectList) {
					reactiveEntity =
						(ReactiveEntity)this.entityToConnectList.remove(0);
				}
				this.connectedEntitySet.add(reactiveEntity);
				if (trace != null && isFullTrace)
					trace.println(this + " " + reactiveEntity + " connected");
			}
			this.activeEntityCount = this.connectedEntitySet.size();
			this.instantHasBeenIdle = this.activeEntityCount == 0;
			if (this.instantHasBeenIdle) {
				this.broadcastDeferredEventList.clear();
				this.isActionMessageGenerated = false;
				this.clearAllAttachments();
				this.status = STATUS_IDLE;
				if (trace != null)
					trace.println(this + " idle");
				if (this.isClosed) {
					this.reactiveMachine.notifyCommunityTermination(this);
					return;
				}
			}
			else {
				this.board = this.futureBoard;
				if (this.isActionMessageGenerated) {
					Object[] attachments = this.actionMessage.getAttachments();
					for (int k = 0; k < attachments.length; k++)
						this.processActionAttachment(
							(ActionAttachmentImpl)attachments[k]);
				}
				this.broadcastInstantaneousEventList.addAll(this.broadcastDeferredEventList);
				this.broadcastDeferredEventList.clear();
				this.isActionMessageGenerated = false;
				while (this.waitingSuspendedEntityCount
						!= this.suspendedEntityCount)
					Thread.yield();
				this.waitingSuspendedEntityCount = 0;
				this.suspendedEntityCount = 0;
				synchronized (this.synchroReleaseSuspend) {
					this.synchroReleaseSuspend.notifyAll();
					this.status = STATUS_ACTIVE;
				}
				synchronized (this.synchroNewInstantActive) {
					this.synchroNewInstantActive.notifyAll();
				}
			}
		}
	}
		
	///////////////////////////
	// Private instance methods
	///////////////////////////
			
	private void processActionAttachment(ActionAttachmentImpl attachment) {
		ActionImpl action = (ActionImpl)attachment.getAction();
		int actionStatus = action.getStatus();
		if (actionStatus == Action.STATUS_PROPOSED)
			action.setStatus(Action.STATUS_EXECUTABLE);
		else if (actionStatus == Action.STATUS_EXECUTABLE) {
			action.setStatus(Action.STATUS_EXECUTED);
			if (trace != null && isFullTrace)
				trace.println(this + " " + attachment.getSenderEntityId()
						+ " executes action");
			try {
				action.execute(this.boardWritingKey);
			}
			catch (Exception exception) {
				attachment.setException(exception);
				if (trace != null) {
					trace.println(this + " "
						+ attachment.getSenderEntityId()
						+ " action " + action + " fails");
					trace.println("  by " + exception);
				}
			}
		}
	}
	
	private void clearAllAttachments() {
		Iterator iterator = this.eventList.iterator();
		while (iterator.hasNext()) {
			Event event = (Event)iterator.next();
			if (event.isMessage())
				((MessageImpl)event).clearAttachments();
		}
	}

	private final void synchronizeAutomata() {
		if (this.isAutomataNotified || !this.isAutomataWaiting)
			Thread.yield();
		while (this.isAutomataNotified || !this.isAutomataWaiting)
			this.reactiveMachine.breathe("CommunityAutomataBeReady");
	}
	
	private final void notifyEntityNoMoreActive(ReactiveEntity reactiveEntity) {
		int activeCount;
		synchronized (this.synchroInactivity) {
			activeCount = this.activeEntityCount - 1;
			this.activeEntityCount = activeCount;
		}
		if (activeCount == 0) {
			this.synchronizeAutomata();
			synchronized (this.synchroAutomataActionRequired) {
				this.isAutomataNotified = true;
				this.synchroAutomataActionRequired.notifyAll();
			}
		}
	}
		
	private void verifyStatusActive() {
		int status = this.status;
		if (status != STATUS_ACTIVE)
			throw new ReactiveException(
					"This ReactiveCommunity is not active (status = "
					+ status + ").");
	}

	private void waitForStatusActive() {
		synchronized (this.synchroNewInstantActive) {
			if (this.status != STATUS_ACTIVE)
				try {
					this.synchroNewInstantActive.wait();
				}
				catch (InterruptedException exception) {
				}
		}
	}
	
	private boolean isThereEventOrConnectionProcessing() {
		return
			!this.broadcastInstantaneousEventList.isEmpty()
			|| !this.broadcastDeferredEventList.isEmpty()
			|| (this.broadcastEvents != null)
			|| (!this.entityToConnectList.isEmpty());
	}
	
	/////////////////////////////////////
	// CommunityFromEntity implementation
	/////////////////////////////////////

	void connect(ReactiveEntity reactiveEntity,
			ConnectionAcknowledgment connectionAcknowledgment) {
		if (this.isClosed)
			throw new ClosedCommunityException();
		synchronized (this.entityToConnectList) {
			this.entityToConnectList.add(reactiveEntity);
		}
		if (connectionAcknowledgment != null)
			connectionAcknowledgment.acknowledgeConnection(
				reactiveEntity.getEntityId());
		synchronized (this.synchroEventOrConnection) {
			this.synchroEventOrConnection.notifyAll();
		}
		boolean wasIdle;
		synchronized (this.synchroStatus) {
			wasIdle = this.status == STATUS_IDLE;
			if (wasIdle)
				this.status = STATUS_BEFORE_ACTIVE;
		}
		if (wasIdle) {
			this.connectedEntitySet.add(reactiveEntity);
			this.entityToConnectList.remove(reactiveEntity);
			if (trace != null && isFullTrace)
				trace.println(this + " " + reactiveEntity + " connected");
			this.activeEntityCount += 1;
			this.status = STATUS_ACTIVE;
		}
		else {
			Thread.yield();
			while (this.entityToConnectList.contains(reactiveEntity))
				this.reactiveMachine.breathe("EntityBeConnected");
			this.waitForStatusActive();
		}
	}

	void generateEvent(ReactiveEntity reactiveEntity, Event event) {
		this.verifyStatusActive();
		if (!event.isBaseEvent())
			throw new ReactiveException("This Event is not a base Event.");
		if (((EventImpl)event).generate(this.instant)) {
			synchronized (this.broadcastInstantaneousEventList) {
				this.broadcastInstantaneousEventList.add(event);
			}
			synchronized (this.synchroEventOrConnection) {
				this.synchroEventOrConnection.notifyAll();
			}
			if (this.reporter != null)
				this.reporter.reportBaseEventGeneration(event);
			if (trace != null)
				trace.println(this + " " + reactiveEntity
						+ " generate " + event);
		}
	}

	void generateMessage(ReactiveEntity reactiveEntity, Message message,
			Object attachment) {
		this.verifyStatusActive();
		MessageImpl messageI = ((MessageImpl)message); 
		if (messageI.generate(this.instant)) {
			messageI.clearAttachments();
			synchronized (this.broadcastDeferredEventList) {
				this.broadcastDeferredEventList.add(message);
			}
			if (message.isActionMessage())
				this.isActionMessageGenerated = true;
			synchronized (this.synchroEventOrConnection) {
				this.synchroEventOrConnection.notifyAll();
			}
			if (this.reporter != null)
				this.reporter.reportMessageGeneration(message);
		}
		if (attachment != null) {
			messageI.addAttachment(attachment);
			if (this.reporter != null)
				this.reporter.reportMessageAttachment(message);
		}
		if (trace != null)
			trace.println(this + " " + reactiveEntity
					+ " generate " + message + " " + attachment);
	}
	
	boolean waitEvent(ReactiveEntity reactiveEntity, int delay_ms) {
		this.verifyStatusActive();
		if (trace != null && isFullTrace)
			trace.println(this + " " + reactiveEntity + " waitEvent");
		if (this.isThereEventOrConnectionProcessing())
			return false;
		try {
			synchronized (this.synchroEventOrConnection) {
				this.synchroEventOrConnection.wait(delay_ms);
			}
		}
		catch (InterruptedException exception) {}
		return !this.isThereEventOrConnectionProcessing();
	}
	
	Event[] suspend(ReactiveEntity reactiveEntity) {
		this.verifyStatusActive();
		if (trace != null && isFullTrace)
			trace.println(this + " " + reactiveEntity + " suspend");
		this.suspendedEntityCount += 1;
		this.notifyEntityNoMoreActive(reactiveEntity);
		synchronized (this.synchroReleaseSuspend) {
			this.waitingSuspendedEntityCount += 1;
			try {
				this.synchroReleaseSuspend.wait();
			}
			catch (InterruptedException exception) {
			}
		}
		return this.broadcastEvents;
	}

	void stop(ReactiveEntity reactiveEntity) {
		this.verifyStatusActive();
		if (trace != null && isFullTrace)
			trace.println(this + " " + reactiveEntity + " stop");
		long instant = this.instant;
		this.notifyEntityNoMoreActive(reactiveEntity);
		if (this.instant == instant)
			synchronized (this.synchroNewInstantActive) {
				try {
					this.synchroNewInstantActive.wait();
				}
				catch (InterruptedException exception) {
				}
			}
		else
			this.synchronizeAutomata();
	}

	void disconnect(ReactiveEntity reactiveEntity) {
		this.verifyStatusActive();
		this.connectedEntitySet.remove(reactiveEntity);
		if (trace != null && isFullTrace)
			trace.println(this + " " + reactiveEntity
					+ " disconnected");
		this.notifyEntityNoMoreActive(reactiveEntity);
	}

	void closeCommunity(ReactiveEntity reactiveEntity) {
		if (reactiveEntity != this.creatorEntity)
			throw new CommunityOwnerException();
		if (this.isClosed)
			return;
		this.isClosed = true;
		if (trace != null)
			trace.println(this
					+ ((reactiveEntity != null) ? " " + reactiveEntity : "")
					+ " close");
		if (this.status == STATUS_IDLE)
			synchronized (this.synchroAutomataActionRequired) {
				this.synchroAutomataActionRequired.notifyAll();
			}
	}

	Event createEvent(ReactiveEntity reactiveEntity) {
		Event event;
		event = new EventImpl(this, reactiveEntity);
		synchronized (this.eventList) {
			this.eventList.add(event);
		}
		return event;
	}

	Message createMessage(ReactiveEntity reactiveEntity, boolean isSenderCertified) {
		Message message;
		message = new MessageImpl(this, reactiveEntity, isSenderCertified, false);
		synchronized (this.eventList) {
			this.eventList.add(message);
		}
		return message;
	}

	void nameEvent(ReactiveEntity reactiveEntity, Event event, String name) {
		boolean isOk = true;
		synchronized (this.fromNameToEvent) {
			if (this.fromNameToEvent.containsKey(name))
				isOk = false;
			else {
				((EventImpl)event).setName(name);
				this.fromNameToEvent.put(name, event);
			}
		}
		if (!isOk)
			throw new NameAlreadyUsedException();
		if (trace != null)
			trace.println(this + " " + reactiveEntity + " " + event
					+ " is \"" + name + "\"");
	}

	void unvalidateEvent(ReactiveEntity reactiveEntity, Event event) {
		synchronized (this.fromNameToEvent) {
			String name = event.getName();
			if (name != null)
				this.fromNameToEvent.remove(name);
			this.eventList.remove(event);
			((EventImpl)event).unvalidate();
		}
	}
	
	///////////////////////////////////
	// ReactiveCommunity implementation
	///////////////////////////////////
	
	public long getInstant() {
		return this.instant;
	}

	public boolean isIdle() {
		return this.status == STATUS_IDLE;
	}
	
	public boolean isClosed() {
		return this.isClosed;
	}

	public Event searchEvent(String name) {
		if (name == null)
			throw new IllegalArgumentException();
		Event event = (Event)this.fromNameToEvent.get(name);
		if (event == null)
			throw new UndefinedEventOrMessageNameException(name);
		return event;
	}

	public Message getActionMessage() {
		return this.actionMessage;
	}

	public Board getBoard() {
		return this.board;
	}
	
	public void setBoard(Board board, long writingKey) {
		if (this.futureBoard != null)
			throw new BoardAlreadyAttachedException();
		if (board.getCommunity() != this)
			throw new BoardAttachmentException();
		this.futureBoard = board;
		this.boardWritingKey = writingKey;
	}
	
	public String getTheme() {
		return this.theme;
	}
	
	////////////////////////
	// Object implementation
	////////////////////////

	public String toString() {
		return this.communityId;
	}
}
