/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.impl;

import java.io.PrintStream;

import org.xmloperator.reactive.exception.BadCharInNameException;
import org.xmloperator.reactive.exception.NameAlreadyUsedException;
import org.xmloperator.reactive.exception.NoMoreValidEventException;
import org.xmloperator.reactive.exception.DisconnectedEntityException;
import org.xmloperator.reactive.exception.NotOwnThreadException;
import org.xmloperator.reactive.exception.EventOwnerException;
import org.xmloperator.reactive.exception.UndefinedEventOrMessageNameException;
import org.xmloperator.reactive.exception.PreventedActionException;

import org.xmloperator.reactive.Action;
import org.xmloperator.reactive.ActionAttachment;
import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.Event;
import org.xmloperator.reactive.CertifiedAttachment;
import org.xmloperator.reactive.LaunchingContext;
import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveEntity;
import org.xmloperator.reactive.ReactiveMachine;

/**
 * ReactiveEntity implementation.
 */
final class ReactiveEntityImpl implements ReactiveEntity {

	private static int reactiveEntityCount = 0;
	private static PrintStream trace = null;
	private static boolean isFullTrace = false;
	
	static void setTrace(PrintStream trace, boolean isFullTrace) {
		ReactiveEntityImpl.trace = trace;
		ReactiveEntityImpl.isFullTrace = isFullTrace;
	}
	
	private static void verifyEventValidity(Event event) {
		if (!event.isValid())
			throw new NoMoreValidEventException(); 
	}

	private final int id = ++ReactiveEntityImpl.reactiveEntityCount;
	private final String entityId = ("A" + String.valueOf(this.id)).intern();
	private final ReactiveMachineImpl machine;
	private final String agentProfile;
	private Thread thread = null;
	private boolean isSenderCertifiedMessageGenerationAllowed;
	private boolean isProposedActionGenerationAllowed;
	private boolean isCommunityConnectionAllowed;
	private boolean isEventGenerationAllowed;
	private boolean isUncertifiedMessageGenerationAllowed;
	private boolean isCommunityCreationAllowed;
	private boolean isExecutableActionGenerationAllowed;
	private long agentLaunchingKey;
	private boolean isAnyAccessPrevented = false;
	private ReactiveCommunityImpl community = null;
	
//	private boolean isEntityWaiting = false;
//	private boolean isEntityNotified = false;

	ReactiveEntityImpl(
			ReactiveMachineImpl machine,
			String agentProfile,
			long agentLaunchingKey) {
		if (machine == null || agentProfile == null)
			throw new IllegalArgumentException();
		this.machine = machine;
		this.agentProfile = agentProfile;
		this.isSenderCertifiedMessageGenerationAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_SENDER_CERTIFIED_MESSAGE_GENERATION)
				>= 0;
		this.isProposedActionGenerationAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_ACTION_PROPOSITION) >= 0;
		this.isCommunityConnectionAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_COMMUNITY_CONNECTION) >= 0;
		this.isEventGenerationAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_EVENT_GENERATION) >= 0;
		this.isUncertifiedMessageGenerationAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_UNCERTIFIED_MESSAGE_GENERATION) >= 0;
		this.isCommunityCreationAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_COMMUNITY_CREATION) >= 0;
		this.isExecutableActionGenerationAllowed = agentProfile.indexOf(
				ReactiveMachine.PROFILE_ACTION_EXECUTION) >= 0;
		this.agentLaunchingKey = agentLaunchingKey;
	}
	
	void setThread(Thread thread) {
		this.thread = thread;
	}
	
	void setCommunity(ReactiveCommunityImpl community) {
		this.community = community;
	}
	
//	final void waitForEvent() {
//		if (this.isEntityWaiting || this.isEntityNotified)
//			throw new RuntimeException("ReactiveEntity not synchronized");
//		synchronized (this) {
//			try {
//				this.isEntityWaiting = true;
//				this.wait();
//			}
//			catch (InterruptedException exception) {
//			}
//		}
//		this.isEntityWaiting = false;
//		this.isEntityNotified = false;
//	}
//	
//	final void notifyEvent() {
//		while (this.isEntityNotified || !this.isEntityWaiting) {
//			Thread.yield();
//		}
//		synchronized (this) {
//			this.isEntityNotified = true;
//			this.notifyAll();
//		}
//	}

	void prevent(boolean forAnyAccess) {
		this.isCommunityConnectionAllowed = false;
		this.isEventGenerationAllowed = false;
		this.isUncertifiedMessageGenerationAllowed = false;
		this.isSenderCertifiedMessageGenerationAllowed = false;
		this.agentLaunchingKey = 0;
		if (forAnyAccess)
			this.isAnyAccessPrevented = true;
	}
	
	private void verifyCurrentThread() {
		if (Thread.currentThread() != this.thread)
			throw new NotOwnThreadException();
	}

	private void verifyConnected() {
		if (this.community == null)
			throw new DisconnectedEntityException();
	}
	
	private void verifyEventOwner(Event event) {
		if (((EventImpl)event).getOwner() != this)
			throw new EventOwnerException();
	}
	
	////////////////////////
	// Object implementation
	////////////////////////

	public String toString() {
		return this.entityId;
	}
	
	////////////////////////////////
	// ReactiveEntity implementation
	////////////////////////////////
	
	public String getEntityId() {
		return this.entityId;
	}
	
	public ReactiveCommunity getCommunity() {
		return this.community;
	}

	public void connect(ReactiveCommunity community) {
		this.verifyCurrentThread();
		if (!this.isCommunityConnectionAllowed)
			throw new PreventedActionException();
		if (this.community != null) {
			if (trace != null && isFullTrace)
				trace.println(this + " disconnect");
			this.community.disconnect(this);
			this.community = null;
		}
		if (trace != null && isFullTrace)
			trace.println(this + " connect");
		ReactiveCommunityImpl communityI = ((ReactiveCommunityImpl)community); 
		communityI.connect(this, null);
		this.community = communityI;
	}

	public void generateEvent(Event event) {
		if (event == null)
			throw new IllegalArgumentException();
		this.verifyCurrentThread();
		this.verifyConnected();
		if (!this.isEventGenerationAllowed)
			throw new PreventedActionException();
		verifyEventValidity(event);
		if (trace != null && isFullTrace)
			trace.println(this + " generate " + event);
		this.community.generateEvent(this, event);
	}
	
	public void generateMessage(Message message, Object attachment) {
		if (message == null)
			throw new IllegalArgumentException();
		this.verifyCurrentThread();
		this.verifyConnected();
		verifyEventValidity(message);
		if (message.isSenderCertified()) {
			if (!this.isSenderCertifiedMessageGenerationAllowed)
				throw new PreventedActionException();
			if (!(attachment instanceof CertifiedAttachment))
				throw new IllegalArgumentException();
			CertifiedAttachmentImpl certifiedAttachment =
				(CertifiedAttachmentImpl)attachment;
			certifiedAttachment.setSenderEntityId(this.entityId);
			if (message.isActionMessage()) {
				if (!certifiedAttachment.isActionAttachment())
					throw new IllegalArgumentException();
				ActionAttachment actionAttachment =
					(ActionAttachment)certifiedAttachment;
				int actionStatus = actionAttachment.getAction().getStatus();
				if ((actionStatus == Action.STATUS_PROPOSED
							&& !this.isProposedActionGenerationAllowed)
						|| (actionStatus == Action.STATUS_EXECUTABLE
							&& !this.isExecutableActionGenerationAllowed))
					throw new PreventedActionException();
			}
		}
		else {
			if (!this.isUncertifiedMessageGenerationAllowed)
				throw new PreventedActionException();
		}
		if (trace != null && isFullTrace)
			trace.println(this + " generate " + message + " " + attachment);
		this.community.generateMessage(this, message, attachment);
	}

	public void generateActionMessage(Action action) {
		this.generateMessage(
				this.community.actionMessage,
				new ActionAttachmentImpl(action));
	}
	
	public boolean waitEvent(int delay_ms) {
		this.verifyCurrentThread();
		this.verifyConnected();
		if (this.isAnyAccessPrevented)
			throw new PreventedActionException();
		if (trace != null && isFullTrace)
			trace.println(this + " waitEvent");
		return this.community.waitEvent(this, delay_ms);
	}
	
	public Event[] suspend() {
		this.verifyCurrentThread();
		this.verifyConnected();
		if (this.isAnyAccessPrevented)
			throw new PreventedActionException();
		if (trace != null && isFullTrace)
			trace.println(this + " suspend");
		Event[] events;
		events = this.community.suspend(this);
		if (trace != null && isFullTrace && events != null)
			trace.println(this + " receive "
					+ ReactiveCommunityImpl.arrayToString(events));
		return events;
	}

	public void stop() {
		this.verifyCurrentThread();
		this.verifyConnected();
		if (this.isAnyAccessPrevented)
			throw new PreventedActionException();
		if (trace != null && isFullTrace)
			trace.println(this + " stop");
		this.community.stop(this);
	}

	public void disconnect() {
		if (this.community == null)
			return;
		this.verifyCurrentThread();
		if (trace != null && isFullTrace)
			trace.println(this + " disconnect");
		this.community.disconnect(this);
		this.community = null;
	}

	public void closeCommunity() {
		this.verifyCurrentThread();
		this.verifyConnected();
		this.community.closeCommunity(this);
	}
	
	public Event createEvent() {
		this.verifyCurrentThread();
		this.verifyConnected();
		return this.community.createEvent(this);
	}

	public Message createMessage(boolean isSenderCertified) {
		this.verifyCurrentThread();
		this.verifyConnected();
		return this.community.createMessage(this, isSenderCertified);
	}

	public void nameEvent(Event event, String name)
			throws NameAlreadyUsedException {
		if (event == null || name == null)
			throw new IllegalArgumentException();
		BadCharInNameException.verifyName(name);
		this.verifyCurrentThread();
		this.verifyConnected();
		this.verifyEventOwner(event);
		verifyEventValidity(event);
		this.community.nameEvent(this, event, name.intern());
	}

	public Event searchEvent(String name) {
		try {
			Event event = this.community.searchEvent(name);
			if (!event.isBaseEvent())
				throw new NameAlreadyUsedException();
			return event;
		}
		catch (UndefinedEventOrMessageNameException exception1) {
			BadCharInNameException.verifyName(name);
			this.verifyCurrentThread();
			this.verifyConnected();
			Event event = this.community.createEvent(this);
			try {
				this.community.nameEvent(this, event, name.intern());
			}
			catch (NameAlreadyUsedException exception2) {
				this.community.unvalidateEvent(this, event);
				event = this.community.searchEvent(name);
				if (!event.isBaseEvent())
					throw new NameAlreadyUsedException();
			}
			return event;
		}
	}

	public Message searchMessage(boolean isSenderCertified, String name) {
		try {
			Event event = this.community.searchEvent(name);
			if (!event.isMessage())
				throw new NameAlreadyUsedException();
			Message message = (Message)event;
			if (message.isSenderCertified() != isSenderCertified)
				throw new NameAlreadyUsedException();
			return message;
		}
		catch (UndefinedEventOrMessageNameException exception1) {
			BadCharInNameException.verifyName(name);
			this.verifyCurrentThread();
			this.verifyConnected();
			Message message = this.community.createMessage(this, isSenderCertified);
			try {
				this.community.nameEvent(this, message, name.intern());
			}
			catch (NameAlreadyUsedException exception2) {
				this.community.unvalidateEvent(this, message);
				Event event = this.community.searchEvent(name);
				if (!event.isMessage())
					throw new NameAlreadyUsedException();
				message = (Message)event;
				if (message.isSenderCertified() != isSenderCertified)
					throw new NameAlreadyUsedException();
			}
			return message;
		}
	}

	public void unvalidateEvent(Event event) {
		if (event == null)
			throw new IllegalArgumentException();
		this.verifyCurrentThread();
		this.verifyConnected();
		this.verifyEventOwner(event);
		if (event.isValid())
			this.community.unvalidateEvent(this, event);
	}

	public ReactiveCommunity newCommunity(String thema) {
		this.verifyCurrentThread();
		if (!this.isCommunityCreationAllowed)
			throw new PreventedActionException();
		return new ReactiveCommunityImpl(this.machine, this, thema);
	}

	public String launchAgent(Agent agent, LaunchingContext context) {
		this.verifyCurrentThread();
		this.verifyConnected();
		if (this.agentLaunchingKey == 0)
			throw new PreventedActionException();
		if (context == null)
			context = new LaunchingContextImpl();
		if (context.getArgumentValue(Agent.ARG_DESTINATION_COMMUNITY) == null)
			context.addArgument(Agent.ARG_DESTINATION_COMMUNITY,
					this.community);
		return this.machine.launchAgent(this.agentLaunchingKey, agent,
				context, this.agentProfile);
	}
}
