/*
 * Copyright (c) 2008 The_xmloperator_project.  All rights reserved.
 * See terms of use on the project's site (http://www.xmloperator.net/reactive/).
 */
package org.xmloperator.reactive.test;

import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.FileOutputStream;

//import org.xml.sax.SAXException;

import org.xmloperator.reactive.exception.ReactiveException;

import org.xmloperator.reactive.Agent;
import org.xmloperator.reactive.LaunchingContext;
//import org.xmloperator.reactive.Message;
import org.xmloperator.reactive.ReactiveCommunity;
import org.xmloperator.reactive.ReactiveMachine;

import org.xmloperator.reactive.facet.AgentCatalogFacet;
import org.xmloperator.reactive.facet.ExitFacet;
import org.xmloperator.reactive.facet.AgentPreventFacet;

import org.xmloperator.reactive.agent.FacetsAgent;

import org.xmloperator.reactive.element.RootElement;
import org.xmloperator.reactive.element.Element;

import org.xmloperator.reactive.element.impl.ElementBoardImpl;
import org.xmloperator.reactive.element.impl.ParentElementImpl;
import org.xmloperator.reactive.element.impl.ObjectDataElementImpl;
import org.xmloperator.reactive.element.impl.StringDataElementImpl;

import org.xmloperator.reactive.element.util.DocumentWriter;
import org.xmloperator.reactive.element.util.DocumentReader;

import org.xmloperator.reactive.element.constraint.DifferentNamesChildrenConstraint;
import org.xmloperator.reactive.element.constraint.FixedNameChildrenConstraint;

//import org.xmloperator.reactive.element.spec.AndElementSpecification;
import org.xmloperator.reactive.element.spec.ChildElementSpecification;
import org.xmloperator.reactive.element.spec.IncludingDescendanceElementSpecification;
//import org.xmloperator.reactive.element.spec.NameElementSpecification;
import org.xmloperator.reactive.element.spec.RootElementSpecification;

import org.xmloperator.reactive.element.action.reject.ElementTouchRejector;

import org.xmloperator.reactive.element.facet.ActionExecutionFacet;

/**
 * A CommunityManagerAgent creates a ReactiveCommunity and manages it
 *   during its full life.
 */
public class CommunityManagerAgent extends FacetsAgent {
	
	/**
	 * Creates, initializes and returns a Board for a ReactiveCommunity.
	 * Board schema :
	 *   board (rejects, agents, projects)
	 *   rejects (reject*)
	 *   agents (entityId[agentId]*)
	 *   projects (*)
	 * @param community a ReactiveCommunity.
	 * @param managerEntityId the entityId of the Agent that manages
	 *   the community.
	 * @param managerAgentId the agentId of the Agent that manages
	 *   the community.
	 * @param writingKey a key for setting the Board.
	 * @return the created Board.
	 */
	protected static ElementBoardImpl createInitialBoard(
			ReactiveCommunity community,
			String managerEntityId,
			String managerAgentId,
			long writingKey) {
		Element rejectsElement =
			new ParentElementImpl(Element.REACTIVE_NAMESPACE, "rejects",
					boardActionRejectors(),
					new FixedNameChildrenConstraint(
							Element.REACTIVE_NAMESPACE, "reject"));
		Element managerAgentElement =
			new StringDataElementImpl(
				Element.REACTIVE_NAMESPACE, managerEntityId, managerAgentId);
		Element agentsElement =
			new ParentElementImpl(Element.REACTIVE_NAMESPACE, "agents",
					new Element[] {managerAgentElement},
					DifferentNamesChildrenConstraint.CHILDREN_CONSTRAINT);
		Element projectsElement =
			new ParentElementImpl(Element.REACTIVE_NAMESPACE, "projects",
			DifferentNamesChildrenConstraint.CHILDREN_CONSTRAINT);
		Element[] boardChildren =
			new Element[] {rejectsElement, agentsElement, projectsElement};
		return new ElementBoardImpl(community, boardChildren, writingKey);
	}
	
	protected static Element[] boardActionRejectors() {
		return new Element[] {
			/*
			 * Do not change children of "board".
			 */
			new ObjectDataElementImpl(
				Element.REACTIVE_NAMESPACE,
				"reject",
				ElementTouchRejector.newModifyRejector(
					RootElementSpecification.ROOT)),
			/*
			 * Do not change children of "rejects".
			 */
			new ObjectDataElementImpl(
				Element.REACTIVE_NAMESPACE,
				"reject",
				ElementTouchRejector.newModifyRejector(
					new ChildElementSpecification(
						RootElementSpecification.ROOT,
						Element.REACTIVE_NAMESPACE,
						"rejects"))),
			/*
			 * Do not change children of "agents".
			 */
			new ObjectDataElementImpl(
				Element.REACTIVE_NAMESPACE,
				"reject",
				ElementTouchRejector.newModifyRejector(
					new IncludingDescendanceElementSpecification(
						new ChildElementSpecification(
							RootElementSpecification.ROOT,
							Element.REACTIVE_NAMESPACE,
							"agents"))))
		};
	}
	
	protected String theme = null;
	protected String boardFileName = null;
	protected long globalExitKey = 0;
	protected long thisAgentExitKey = 0;
	protected ExitFacet exitFacet = null;
	
	public CommunityManagerAgent() {
	}
	
	public CommunityManagerAgent(String theme, String boardFileName,
			long globalExitKey, long thisAgentExitKey) {
		this();
		if (theme != null)
			this.setThema(theme);
		if (boardFileName != null)
			this.setBoardFileName(boardFileName);
		this.setExitKeys(globalExitKey, thisAgentExitKey);
	}
	
	public void setThema(String theme) {
		if (theme == null)
			throw new IllegalArgumentException();
		if (this.theme != null)
			throw new IllegalStateException();
		this.theme = theme.intern();
	}
	
	public void setBoardFileName(String boardFileName) {
		if (boardFileName == null)
			throw new IllegalArgumentException();
		if (this.boardFileName != null)
			throw new IllegalStateException();
		this.boardFileName = boardFileName;
	}
	
	public void setExitKeys(long globalExitKey, long thisAgentExitKey) {
		this.globalExitKey = globalExitKey;
		this.thisAgentExitKey = thisAgentExitKey;
		if (this.exitFacet != null)
			this.exitFacet.setExitKeys(globalExitKey, thisAgentExitKey);
	}
	
	protected ReactiveCommunity newCommunity() {
		ReactiveMachine reactiveMachine =
			(ReactiveMachine)this.launchingContext.getArgumentValue(
					Agent.ARG_REACTIVE_MACHINE);
		if (this.theme != null
				&& reactiveMachine.getCommunity(this.theme) != null)
			throw new ReactiveException("This theme already exists");
		ReactiveCommunity community = this.reac.newCommunity(this.theme);
		if (this.trace != null)
			this.trace.println(community
				+ (this.theme != null
					? " theme is \"" + this.theme + "\""
					: " has no theme"));
		return community;
	}
	
	///////////////////////
	// Agent implementation
	///////////////////////

	public void init(LaunchingContext launchingContext) {
		super.init(launchingContext);
		this.exitFacet = new ExitFacet();
		this.exitFacet.setExitKeys(this.globalExitKey, this.thisAgentExitKey);
		this.addFacet(this.exitFacet);
		this.addFacet(new AgentPreventFacet());
		this.addFacet(new AgentCatalogFacet());
		this.addFacet(new ActionExecutionFacet());
		if (this.boardFileName != null) {
			ReactiveCommunity community = this.reac.getCommunity();
			ReactiveMachine reactiveMachine =
				(ReactiveMachine)launchingContext.getArgumentValue(
					Agent.ARG_REACTIVE_MACHINE);
			long writingKey = reactiveMachine.newKey();
			ElementBoardImpl board =
				new ElementBoardImpl(community, writingKey);
			try {
				DocumentReader.readDocument(
					new FileInputStream(this.boardFileName),
					board,
					board,
					writingKey);
				if (this.trace != null)
					this.trace.println(community + " board read on \""
							+ this.boardFileName + "\"");
			}
			catch (Exception exception) {
				exception.printStackTrace();
				board = null;
			}
			if (board == null) {
				board =
					createInitialBoard(
						community,
						this.reac.getEntityId(),
						this.agentId,
						writingKey);
				if (this.trace != null)
					this.trace.println(community + " board initialized");
			}
			community.setBoard(board, writingKey);
		}
	}
	
	public boolean continues() {
		return !this.exitFacet.isExit();
	}

	public void terminate() {
		super.terminate();
		ReactiveCommunity community = this.reac.getCommunity();
		this.reac.closeCommunity();
		this.reac.disconnect();
		Thread.yield();
		ReactiveMachine reactiveMachine =
			(ReactiveMachine)this.launchingContext.getArgumentValue(
					Agent.ARG_REACTIVE_MACHINE);
		while (!community.isIdle())
			reactiveMachine.breathe("CommunityBeIdle");
		if (this.boardFileName != null) {
			 RootElement documentElement =
				 (ElementBoardImpl)community.getBoard();
			if (documentElement != null)
				try {
					DocumentWriter.printDocument(
							new FileOutputStream(this.boardFileName),
							documentElement,
							null);
					if (this.trace != null)
						this.trace.println(community + " board saved on \""
								+ this.boardFileName + "\"");
				 }
				 catch (FileNotFoundException exception) {
				 	throw new ReactiveException(
						"Unable to save the board on "
						+ this.boardFileName);
				 }
		}
	}
}
